import os
import uuid

from fastapi import Header, Request
from fastapi.exceptions import HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse, RedirectResponse, PlainTextResponse
from nicegui import ui, app, Client
from nicegui.events import GenericEventArguments

from dorky import assets_dir
from dorky.content import operators, help, popup, elements
from dorky.search_engines import Search


def main(port: int):
    # Load the serach engines
    engines_list = ["Google", "Bing", "Yandex", "DuckDuckGo", "Brave", "Baidu"]
    # Assets declaration
    app.add_static_files("assets", assets_dir)

    # Startup message
    app.on_startup(lambda: print(f"{elements.dorky_ascii}\n🚀 Dorky is up and running 🚀\n\n"
                                 f"🌐 Visit http://127.0.0.1:{port} or http://localhost:{port}.\n\n"
                                 f"🛑 Press Ctrl+C to stop Dorky."))

    # Main page
    @ui.page("/", title="Dorky search")
    async def search_page():
        """Main page of Dorky."""
        await ui.context.client.connected(timeout=10)

        # Set default storage values
        for key, default in [
            ("search", ""),
            ("popup_allowed", False),
            ("view_warning", True),
            ("view_help", True),
            ("view_operators", True),
            ("Google", True),
            ("Bing", True),
            ("Yandex", True),
            ("DuckDuckGo", False),
            ("Brave", False),
            ("Baidu", False)
        ]:
            app.storage.user.setdefault(key, default)

        # Add the header
        elements.header()

        # Add the search engines buttons
        with ui.row():
            def create_browser_button(engine_name):
                """Create a button for each search engine_name."""
                with ui.row().style("cursor: pointer;").style(
                        "filter: grayscale(0%);"
                        if app.storage.user[engine_name]
                        else "filter: grayscale(100%);"
                ).on("click", lambda e: image_change(e, engine_name)):
                    ui.image(assets_dir + f"/images/{engine_name.lower()}.png").style(
                        "width: 20px; height: 20px; margin: 0px; padding: 0px; "
                    )
                    ui.label(engine_name).style("color: #0060df; margin: 0 0 0 -10px;")

            for engine in engines_list:
                create_browser_button(engine)

        def trim_search_value(e: GenericEventArguments) -> None:
            """Trim the search input value to remove the new line induced by hitting Enter."""
            if e.sender.value:
                e.sender.value = e.sender.value.strip()

        def _enter_down(e: GenericEventArguments) -> None:
            """Handle the Enter key down event to run the search."""
            e.sender.clear()
            trim_search_value(e)
            run_search()

        # Add the search input and button
        with ui.row().style("width:100%"):
            ui.input(placeholder="Type your search").style("width:80%").props(
                "outlined dense clearable width=100% autogrow"
            ).bind_value(app.storage.user, "search").on("keydown.enter", _enter_down)

            ui.button("Search").on_click(
                run_search
            )

        # Add the help block
        with ui.column():
            help.help_block()
        # Add the operators table
        with ui.row():
            operators.operators_table()
        # Add the footer
        elements.footer()

        # Check if the pop-ups are allowed in the browser
        await popup.check(port)

    # Add the CORS header for the API endpoint cerating a FastAPI sub app
    api_app = app

    api_app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=False,
        allow_methods=["*"],
        allow_headers=["X-Dorky"]
    )

    # Create the API endpoint to use with the extension
    @api_app.get("/api/")
    def api_get_converted_urls(engines: str = "", search: str = "", source: str = "", x_dorky: str = Header(None)) -> JSONResponse or RedirectResponse:
        """API endpoint for the ff addon to convert the search query to URLs."""
        if x_dorky is None:
            return RedirectResponse(url="/")
        # Return an error if the search is empty
        if search == "":
            return JSONResponse(
                status_code=400,
                content={
                    "status": "error"
                },
            )
        # Convert the search query to URLs using the selected engines
        engines = engines.split(",")
        result = Search(search, engines).convert()
        if result["status"] == "ok":
            # Reverse the URLs as the add-on behaviour is different from the web app when opening the tabs
            result["urls"].reverse()

        return JSONResponse(content=result)

    @ui.page("/popup", title="Dorky Popup Check")
    def popup_check_page():
        """Popup check page for Dorky."""
        ui.label("Popup check page for Dorky.").style("text-align: center; margin-top: 20px;")
        ui.label("You can close this window.").style("text-align: center; margin-top: 10px;")

    # Toggle the grayscale filter for the chosen search engine images
    def image_change(e: GenericEventArguments, engine: str) -> None:
        """Toggle the grayscale filter for the chosen search engine images."""
        if app.storage.user[engine]:
            app.storage.user[engine] = False
            e.sender.style("filter: grayscale(100%)")
        else:
            app.storage.user[engine] = True
            e.sender.style("filter: grayscale(0%)")

    def run_search() -> None:
        """Run the convertor with the selected engines and open tabs."""
        if not app.storage.user["search"] or app.storage.user["search"].strip() == "":
            ui.notify("Fill the search input with your query.", position="center", type="negative", close_button=True)
            return
        engines = [
            key
            for key, value in app.storage.user.items()
            if value and key in engines_list
        ]
        if not engines:
            ui.notify("No search engines selected. Please select at least one engine.", position="center", type="negative", close_button=True)
            return
        search_urls = Search(app.storage.user["search"], engines).convert()
        if search_urls["status"] == "error":
            ui.notify(search_urls["error"], position="center", type="negative", close_button=True)
        else:
            for url in search_urls["urls"]:
                ui.navigate.to(url, new_tab=True)

    @app.exception_handler(404)
    async def not_found_exception_handler(request: Request, exc: HTTPException) -> RedirectResponse:
        """Handle 404 errors by redirecting to the main page."""
        return RedirectResponse("/")

    @app.exception_handler(500)
    async def not_found_exception_handler(request: Request, exc: HTTPException) -> None:
        return

    @app.get("/robots.txt", response_class=PlainTextResponse)
    def robots() -> str:
        """Serve the robots.txt file."""
        with open(assets_dir + "/robots.txt", "r") as file:
            return file.read()

    # Check if the storage secret is set with the env variable;
    # if not, delete the storage and set a new secret.
    if secret := os.getenv("DORKY_SECRET"):
        print(f"🔑 Using DORKY_SECRET from environment for storage.")
    else:
        app.storage.clear()
        print(f"🗑️DORKY_SECRET not found, storage deleted.")
        secret = str(uuid.uuid4())
        print(f"🔑 New secret generated for storage.")

    # Run the app
    ui.run(
        port=port,
        storage_secret=secret,
        show=False,
        reload=False,
        show_welcome_message=False,
        favicon=assets_dir + "/images/favicon.ico",
    )



from functools import partial

from nicegui import ui, app

from dorky import assets_dir


async def check(port: int) -> None:
    """Check if pop-ups are allowed in the browser and show a warning dialog if not."""
    async def popup_state() -> bool:
        """Check if pop-ups are allowed in the browser by opening a small blank popup."""
        is_allowed = await ui.run_javascript(
            f"""
            (() => {{
                const p = window.open('http://127.0.0.1:{port}/popup', '', 'width=300,height=150');
                if (!p) return false;
                p.close();
                return true;
            }})()
            """,
            timeout=15,
        )
        return is_allowed

    # Does thee popup state has been checked before?
    popup_allowed = app.storage.user.get("popup_allowed")
    # Does the alert for the user has been already shown?
    view_warning = app.storage.user.get("view_warning")
    # Check as sessions do expire
    if not popup_allowed and view_warning:
        # Get the browser popup state
        popup_state = await popup_state()
        # If popups are allowed, update the storage
        if popup_state:
            app.storage.user.update(popup_allowed=True)
            app.storage.user.update(view_warning=False)
        else:
            # Show the warning dialog if pop-ups are not allowed
            with ui.dialog() as dialog, ui.card():
                # Do not show the warning again
                app.storage.user.update(view_warning=False)
                with ui.row().style():
                    ui.image(assets_dir + "/images/dorky-head.png").style(
                        "width: 30px; height: 30px; margin: 0px; padding: 0px;"
                    ).props("fit=contain")
                    ui.label("Allow pop-ups for dorky domain").classes("text-h6")
                ui.label(
                    "Pop-up windows should be allowed for Dorky website to work properly."
                )
                with ui.row().style(
                    "width: 100%; display: flex; align-items: center; justify-content: space-between;"
                ):
                    ui.link("Get some help", "https://support.mozilla.org/en-US/kb/pop-blocker-settings-exceptions-troubleshooting#w_what-are-pop-ups", new_tab=True)
                    ui.button("Close", on_click=dialog.close)

            # Show the dialog
            async def show():
                await dialog

            # Run the dialog only once without user interaction
            ui.timer(0, partial(show), once=True)



from nicegui import ui, app

from dorky.content import elements

operator_list = [
    {
        "table": {
            "operator": '" "',
            "name": "Quotation marks",
            "native": "yes",
            "google": "yes",
            "bing": "yes",
            "yandex": "yes",
            "duckduckgo": "yes",
            "brave": "yes",
            "baidu": "yes",
        },
        "operator": '" "',
        "name": "Quotations marks",
        "description": "Results that mention an exact word or phrase",
        "examples": [
            {
                "example": '"trace labs global"',
                "explanation": "Results containing the exact phrase <code>trace labs global</code>.",
            }
        ],
    },
    {
        "table": {
            "operator": "( )",
            "name": "Parentheses",
            "native": "yes",
            "google": "yes",
            "bing": "yes",
            "yandex": "yes",
            "duckduckgo": "yes",
            "brave": "yes",
            "baidu": "yes",
        },
        "description": "Group multiple operators",
        "examples": [
            {
                "example": "osint AND (geoint OR socmint)",
                "explanation": "Results containing <code>osint</code> and either <code>geoint</code> or <code>socmint</code>.",
            }
        ],
    },
    {
        "table": {
            "operator": "OR, |",
            "name": "OR operator",
            "native": "yes",
            "google": "yes",
            "bing": "yes",
            "yandex": "yes",
            "duckduckgo": "yes",
            "brave": "yes",
            "baidu": "yes",
        },
        "description": "Search for X or Y",
        "examples": [
            {
                "example": "osint OR geoint",
                "explanation": "Results containing either <code>osint</code> or <code>geoint</code>.",
            },
            {
                "example": "osint | geoint",
                "explanation": "Results containing either <code>osint</code> or <code>geoint</code>.",
            },
        ],
    },
    {
        "table": {
            "operator": "AND",
            "name": "AND operator",
            "native": "yes",
            "google": "yes",
            "bing": "yes",
            "yandex": "yes",
            "duckduckgo": "yes",
            "brave": "yes",
            "baidu": "yes",
        },
        "description": "Search for X and Y",
        "examples": [
            {
                "example": "osint AND geoint",
                "explanation": "Results containing both <code>osint</code> and <code>geoint</code>.",
            }
        ],
    },
    {
        "table": {
            "operator": "-",
            "name": "Exclude operator",
            "native": "yes",
            "google": "yes",
            "bing": "yes",
            "yandex": "partial",
            "duckduckgo": "yes",
            "brave": "partial",
            "baidu": "partial",
        },
        "description": "Search without including a specific term.",
        "examples": [
            {
                "example": "osint -geoint",
                "explanation": "Results containing <code>osint</code> but not <code>geoint</code>.",
            },
            {
                "example": "osint -site:github.com",
                "explanation": "Results containing <code>osint</code> but excluding those from <code>github.com</code>.",
            },
        ],
        "notes": [
            {
                "engine": "yandex",
                "note": "Works only with a single word, not with operators. The operator and its value are removed.",
            },
            {
                "engine": "brave",
                "note": "Works only with a single word, not with operators. The operator and its value are removed.",
            },
            {
                "engine": "baidu",
                "note": "Works only with a single word, not with operators. The operator and its value are removed.",
            },
        ],
    },
    {
        "table": {
            "operator": "site:",
            "name": "Site search",
            "native": "yes",
            "google": "yes",
            "bing": "yes",
            "yandex": "yes",
            "duckduckgo": "yes",
            "brave": "yes",
            "baidu": "yes",
        },
        "description": "Search within a specific website",
        "examples": [
            {
                "example": "osint site:github.com",
                "explanation": "Results containing <code>osint</code> from <code>github.com</code>.",
            }
        ],
    },
    {
        "table": {
            "operator": "inurl:",
            "name": "URL search",
            "native": "yes",
            "google": "yes",
            "bing": "no",
            "yandex": "yes",
            "duckduckgo": "yes",
            "brave": "yes",
            "baidu": "yes",
        },
        "description": "Search for a word in the URL",
        "examples": [
            {
                "example": "osint inurl:usa",
                "explanation": "Results containing <code>osint</code> and the word <code>usa</code> in the URL.",
            }
        ],
        "notes": [
            {
                "engine": "bing",
                "note": "No equivalent exists; the operator is removed from the search query, but its value is retained.",
            }
        ],
    },
    {
        "table": {
            "operator": "intitle:",
            "name": "Title search",
            "native": "yes",
            "google": "yes",
            "bing": "yes",
            "yandex": "yes",
            "duckduckgo": "yes",
            "brave": "yes",
            "baidu": "yes",
        },
        "description": "Search for a word in the title",
        "examples": [
            {
                "example": "olympic games intitle:osint",
                "explanation": "Results with <code>olympic games</code> and <code>osint</code> in the title.",
            }
        ],
    },
    {
        "table": {
            "operator": "intext:",
            "name": "Text search",
            "native": "yes",
            "google": "yes",
            "bing": "yes",
            "yandex": "yes",
            "duckduckgo": "no",
            "brave": "yes",
            "baidu": "no",
        },
        "description": "Search for a word in the text or body",
        "examples": [
            {
                "example": "osint intext:tools",
                "explanation": "Results containing both <code>osint</code> and <code>tools</code> in the text.",
            }
        ],
        # TODO: Maybe replacing intext with + or "" ?
        "notes": [
            {
                "engine": "duckduckgo",
                "note": "No equivalent exists; the operator is removed from the search query, but its value is retained.",
            },
            {
                "engine": "baidu",
                "note": "No equivalent exists; the operator is removed from the search query, but its value is retained.",
            },
        ],
    },
    {
        "table": {
            "operator": "filetype:, ext:",
            "name": "File search",
            "native": "yes",
            "google": "yes",
            "bing": "yes",
            "yandex": "yes",
            "duckduckgo": "yes",
            "brave": "yes",
            "baidu": "yes",
        },
        "description": "Search for a specific file type",
        "examples": [
            {
                "example": "osint filetype:pdf",
                "explanation": "Results containing <code>osint</code> in PDF format.",
            },
            {
                "example": "osint ext:pdf",
                "explanation": "Results containing <code>osint</code> in PDF format.",
            },
        ],
    },
    {
        "table": {
            "operator": "*",
            "name": "Wildcard search",
            "native": "yes",
            "google": "yes",
            "bing": "no",
            "yandex": "yes",
            "duckduckgo": "no",
            "brave": "no",
            "baidu": "no",
        },
        # TODO: Check if it's working in other operators like site, intext,...
        "description": 'Search for terms using a wildcard; it only works within <code>""</code>.',
        "examples": [
            {
                "example": '"osint * tools"',
                "explanation": "Results containing <code>osint someword(s) tools</code>.",
            }
        ],
        "notes": [
            {
                "engine": "bing",
                "note": "No equivalent exists; the operator is removed, and the words are placed inside <code>""</code>.",
            },
            {
                "engine": "duckduckgo",
                "note": "No equivalent exists; the operator is removed, and the words are placed inside <code>""</code>.",
            },
            {
                "engine": "brave",
                "note": "No equivalent exists; the operator is removed, and the words are placed inside <code>""</code>.",
            },
            {
                "engine": "baidu",
                "note": "No equivalent exists; the operator is removed, and the words are placed inside <code>""</code>.",
            },
        ],
    },
    {
        "table": {
            "operator": "after:",
            "name": "Date search",
            "native": "yes",
            "google": "yes",
            "bing": "yes",
            "yandex": "yes",
            "duckduckgo": "yes",
            "brave": "yes",
            "baidu": "yes",
        },
        "description": "Search for a term starting from a specific date ",
        "examples": [
            {
                "example": "osint after:2025-01-31",
                "explanation": "Results containing <code>osint</code> from January 31, 2025.",
            }
        ],
    },
    {
        "table": {
            "operator": "before:",
            "name": "Date search",
            "native": "yes",
            "google": "yes",
            "bing": "yes",
            "yandex": "yes",
            "duckduckgo": "yes",
            "brave": "yes",
            "baidu": "yes",
        },
        "description": "Search for a term up to a specific date",
        "examples": [
            {
                "example": "osint before:2024-12-31",
                "explanation": "Results containing <code>osint</code> up to December 31, 2024.",
            }
        ],
    },
    {
        "table": {
            "operator": "lang:",
            "name": "Language search",
            "native": "no",
            "google": "yes",
            "bing": "yes",
            "yandex": "yes",
            "duckduckgo": "no",
            "brave": "partial",
            "baidu": "no",
        },
        "description": "Search in a specific language using <a href='https://en.wikipedia.org/wiki/List_of_ISO_639_language_codes#Table' style='color: #5898d4' target='_blank'>ISO 639-1 language codes</a>.",
        "examples": [
            {
                "example": "osint lang:es",
                "explanation": "Results that contain <code>osint</code> in Spanish.",
            }
        ],
        "notes": [
            {
                "engine": "duckduckgo",
                "note": "Language is filtered by country, so searching in English could require opening 13 windows (e.g., UK, NZ, ZA, CA, etc.). The operator and its value are removed.",
            },
            {
                "engine": "brave",
                "note": "Equivalent exists but results are limited.",
            },
            {
                "engine": "baidu",
                "note": "Only three languages are supported: all languages, Simplified Chinese, and Traditional Chinese. The operator and its value are removed.",
            },
        ],
    },
    {
        "table": {
            "operator": "country:",
            "name": "Country search",
            "native": "no",
            "google": "yes",
            "bing": "yes",
            "yandex": "yes",
            "duckduckgo": "partial",
            "brave": "yes",
            "baidu": "no",
        },
        "description": "Search within a specific country using <a href='https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2#Officially_assigned_code_elements' style='color: #5898d4' target='_blank'>ISO 3166-1 alpha-2 country codes</a>.",
        "examples": [
            {
                "example": "osint country:us",
                "explanation": "Search results for <code>osint</code> specific to the United States.",
            }
        ],
        "notes": [
            {
                "engine": "duckduckgo",
                "note": "Country options are limited. If a specific country is unavailable, the search will default to a global scope. Since results are filtered by language per country, Dorky can open multiple windows—for example, both 'fr' and 'nl' for Belgium.",
            },
            {
                "engine": "baidu",
                "note": "There is no equivalent; both the operator and its value are removed.",
            },
        ],
    },
    {
        "table": {
            "operator": "sub:",
            "name": "Subdomain search",
            "native": "no",
            "google": "yes",
            "bing": "no",
            "yandex": "yes",
            "duckduckgo": "no",
            "brave": "no",
            "baidu": "no",
        },
        "description": "Search across subdomains of a domain",
        "examples": [
            {
                "example": "sub:tesla.com incident",
                "explanation": "Results about <code>incident</code> in the subdomains of <code>tesla.com</code>..",
            },
            {
                "example": "sub:tesla.com",
                "explanation": "All <code>tesla.com</code> results found in URLs of its subdomains.",
            },
        ],
        "notes": [
            {
                "note": "This special operator only works with Google and Yandex. It will only launch them despite your search engine choice.",
            },
            {
                "note": "When used without other operators, it can find sub-domains that are referenced by search engines.",
            },
        ],
    },
    {
        "table": {
            "operator": "last:",
            "name": "Last duration search",
            "native": "no",
            "google": "yes",
            "bing": "no",
            "yandex": "no",
            "duckduckgo": "no",
            "brave": "no",
            "baidu": "no"
        },
        "description": "Search results from the last X hours (h), minutes (m), or seconds (s) indexed by Google.",
        "examples": [
            {
                "example": "osint last:1h",
                "explanation": "Results about <code>osint</code> indexed by Google within the last hour.",
            },
            {
                "example": "osint last:5m",
                "explanation": "Results about <code>osint</code> indexed by Google within the last 5 minutes.",
            },
            {
                "example": "osint last:30s",
                "explanation": "Results about <code>osint</code> indexed by Google within the last 30 seconds.",
            }
        ],
        "notes": [
            {
                "note": "This operator is exclusive to Google and will force a Google search, no matter which search engine you use.",
            },
            {
                "note": "It can't be used with <code>before:</code> or <code>after:</code>; both are ignored.",
            },
            {
                "note": "Using minutes or seconds can help monitor a hot topic in real time.",
            }
        ]
    },
    {
        "table": {
            "operator": "!IP",
            "name": "Search in IPs",
            "native": "no",
            "google": "yes",
            "bing": "no",
            "yandex": "no",
            "duckduckgo": "no",
            "brave": "no",
            "baidu": "no"
        },
        "description": "Search in the IPs indexed by Google, not domains.",
        "examples": [
            {
                "example": "osint !IP",
                "explanation": "Results in the IPs indexed by Google about <code>osint</code>.",
            }
        ],
        "notes": [
            {
                "note": "This operator is exclusive to Google and will force a Google search, no matter which search engine you use.",
            },
            {
                "note": "It opens multiple tabs to cover the entire IPv4 range, depending on the length of your query. It can trigger Google reCAPTCHAs.",
            },
            {
                "note": "Adapted from <a href='https://github.com/SeifElsallamy/gip' target='_blank'>gip</a> by Seif Elsallamy.",
            }
        ]
    }
]

columns = [
    {"name": "google", "label": "Google", "field": "google"},
    {"name": "bing", "label": "Bing", "field": "bing"},
    {"name": "yandex", "label": "Yandex", "field": "yandex"},
    {"name": "duckduckgo", "label": "DuckDuckGo", "field": "duckduckgo"},
    {"name": "brave", "label": "Brave", "field": "brave"},
    {"name": "baidu", "label": "Baidu", "field": "baidu"},
]

column_defaults = {
    "align": "center",
    "headerClasses": "uppercase text-primary",
    "style": "width: 80px",
}


def table_emotes(table):
    table.add_slot(
        "body-cell-native",
        """
        <q-td key="native" :props="props">
            <q-icon
                :name="props.value === 'yes' ? 'check' : props.value === 'no' ? 'close' : props.value === 'partial' ? 'sentiment_satisfied' : 'help_outline'"
                :color="props.value === 'yes' ? 'green' : props.value === 'no' ? 'red' : props.value === 'partial' ? 'orange' : 'grey'"
                size="md"
             />
        </q-td>
    """,
    )
    table.add_slot(
        "body-cell-google",
        """
        <q-td key="google" :props="props">
            <q-icon
                :name="props.value === 'yes' ? 'sentiment_very_satisfied' : props.value === 'no' ? 'sentiment_very_dissatisfied' : props.value === 'partial' ? 'sentiment_satisfied' : 'help_outline'"
                :color="props.value === 'yes' ? 'green' : props.value === 'no' ? 'red' : props.value === 'partial' ? 'orange' : 'grey'"
                size="md"
             />
        </q-td>
    """,
    )
    table.add_slot(
        "body-cell-bing",
        """
        <q-td key="bing" :props="props">
            <q-icon
                :name="props.value === 'yes' ? 'sentiment_very_satisfied' : props.value === 'no' ? 'sentiment_very_dissatisfied' : props.value === 'partial' ? 'sentiment_satisfied' : 'help_outline'"
                :color="props.value === 'yes' ? 'green' : props.value === 'no' ? 'red' : props.value === 'partial' ? 'orange' : 'grey'"
                size="md"
             />
        </q-td>
    """,
    )
    table.add_slot(
        "body-cell-yandex",
        """
        <q-td key="yandex" :props="props">
            <q-icon
                :name="props.value === 'yes' ? 'sentiment_very_satisfied' : props.value === 'no' ? 'sentiment_very_dissatisfied' : props.value === 'partial' ? 'sentiment_satisfied' : 'help_outline'"
                :color="props.value === 'yes' ? 'green' : props.value === 'no' ? 'red' : props.value === 'partial' ? 'orange' : 'grey'"
                size="md"
             />
        </q-td>
    """,
    )
    table.add_slot(
        "body-cell-duckduckgo",
        """
        <q-td key="duckduckgo" :props="props">
            <q-icon
                :name="props.value === 'yes' ? 'sentiment_very_satisfied' : props.value === 'no' ? 'sentiment_very_dissatisfied' : props.value === 'partial' ? 'sentiment_satisfied' : 'help_outline'"
                :color="props.value === 'yes' ? 'green' : props.value === 'no' ? 'red' : props.value === 'partial' ? 'orange' : 'grey'"
                size="md"
             />
        </q-td>
    """,
    )
    table.add_slot(
        "body-cell-brave",
        """
        <q-td key="brave" :props="props">
            <q-icon
                :name="props.value === 'yes' ? 'sentiment_very_satisfied' : props.value === 'no' ? 'sentiment_very_dissatisfied' : props.value === 'partial' ? 'sentiment_satisfied' : 'help_outline'"
                :color="props.value === 'yes' ? 'green' : props.value === 'no' ? 'red' : props.value === 'partial' ? 'orange' : 'grey'"
                size="md"
             />
        </q-td>
    """,
    )
    table.add_slot(
        "body-cell-baidu",
        """
        <q-td key="baidu" :props="props">
            <q-icon
                :name="props.value === 'yes' ? 'sentiment_very_satisfied' : props.value === 'no' ? 'sentiment_very_dissatisfied' : props.value === 'partial' ? 'sentiment_satisfied' : 'help_outline'"
                :color="props.value === 'yes' ? 'green' : props.value === 'no' ? 'red' : props.value === 'partial' ? 'orange' : 'grey'"
                size="md"
             />
        </q-td>
    """,
    )


def operators_table() -> None:
    """Build and display the operators table."""
    with ui.column().style("width:600px"):
        with ui.row():
            elements.ToggleButton("Operators", color="#e6e6e6").on_click(
                lambda: app.storage.user.update(
                    view_operators=(
                        False if app.storage.user["view_operators"] is True else True
                    )
                )
            )
        with ui.row().bind_visibility_from(app.storage.user, "view_operators"):
            with ui.column().style("color: #555;"):
                ui.label("Compatible operators with Dorky.")
                ui.html(
                    "They are mainly Google search operators, but few of them are custom made like <code>lang:</code>, <code>country:</code>, <code>sub:</code> or <code>last:</code>.",
                )

            with ui.row().style("color: #555;"):
                for operator in operator_list:
                    with ui.card().classes("no-shadow border-[1px]").style(
                        "margin: 0px; padding: 0; width:600px"
                    ):
                        op_rows = [operator["table"]]
                        with ui.row().style(
                            "align-items: center; margin:0px; padding:5px 10px; background-color: #f5f5f5; width:100%;"
                        ):
                            ui.label(operator["table"]["operator"]).style(
                                "font-size: 16px; font-weight: bold; color: #3b82f6"
                            )
                            ui.html(operator["description"]).style(
                                "font-size: 14px; color: #15141A;"
                            )

                        with ui.column().style("margin:0px; padding:0 10px 10px 10px;"):
                            table = (
                                ui.table(
                                    columns=columns,
                                    column_defaults=column_defaults,
                                    rows=op_rows,
                                    row_key="name",
                                )
                                .props("flat bordered dense")
                                .style("margin-left: 40px;")
                            )
                            table_emotes(table)

                            for example in operator["examples"]:
                                with ui.row().style(
                                    "margin:0px; padding:0px; align-items:center;"
                                ):
                                    ui.label(example["example"]).style(
                                        "font-size:13px; padding:9px; border:1px solid #ededf0; border-radius:4px;"
                                    ).props()
                                    ui.html(example["explanation"]).style(
                                        "font-size: 14px; margin:0px; padding:0px;"
                                    )

                            if notes := operator.get("notes"):
                                for note in notes:
                                    with ui.row().style(
                                        "margin:0px; padding:0px; align-items:center;"
                                    ):
                                        message = ""
                                        if note.get("engine"):
                                            if (
                                                operator["table"][note["engine"]]
                                                == "no"
                                            ):
                                                color = "#f44336"
                                            elif (
                                                operator["table"][note["engine"]]
                                                == "partial"
                                            ):
                                                color = "#ff9900"
                                            message += f'<span style="color: {color};">{note['engine'].capitalize()}</span>: '
                                        message += note["note"]
                                        ui.html(message)



import argparse
import dorky.dorky as dorky
from dorky import __about__ as about


def main():
    """Main cli function to start Dorky."""
    parser = argparse.ArgumentParser(description="Start Dorky")
    parser.add_argument("start", help="Start Dorky")
    # add port option
    parser.add_argument(
        "-p", "--port", type=int, default=8080, help="Port to run Dorky on (default: 8080)",
    )
    args = parser.parse_args()
    if args.start:
        print(f"\n⚙️ Starting Dorky v{about.__version__} on port {args.port}...")
        dorky.main(port=args.port)



import json
import re
import urllib.parse
from datetime import datetime

from dorky import json_data_dir


class Search:
    """Class to convert search queries"""
    def __init__(self, query, query_engines: list):
        self.query = query
        self.query_engines = query_engines
        self.engines = self.load_engines()
        self.wildcard_pattern = re.compile(r'"[^"]*\*[^"]*"')

    def get_engines_conf(self) -> list:
        """Load the search engines configuration based on the chosen engines."""
        with open(f"{json_data_dir}/search_engines.json", "r", encoding="utf-8") as f:
            engines_conf = json.load(f)
        return [
            engine for engine in engines_conf if engine["name"] in self.query_engines
        ]

    @staticmethod
    def get_yandex_countries() -> list:
        """Load Yandex country codes;"""
        with open(f"{json_data_dir}/yandex_countries.json", "r", encoding="utf-8") as f:
            return json.load(f)

    def load_engines(self) -> list:
        """Get the search engines configuration."""
        engines_conf = self.get_engines_conf()
        engine_set = self.query_engines
        # Engine override for special operators
        if "sub:" in self.query:
            engine_set = ["Google", "Yandex"]
        if "last:" in self.query or "!IP" in self.query:
            engine_set = ["Google"]
        # Retrieve the configuration for the selected search engines
        engines = [engine for engine in engines_conf if engine["name"] in engine_set]
        return engines

    @staticmethod
    def delete_keyword(query, match_group) -> str:
        """Delete a keyword from the query"""
        return query.replace(match_group, "")

    @staticmethod
    def delete_keywords(query, keywords: list, space=False) -> str:
        """Delete keywords from the query"""
        for keyword in keywords:
            try:
                token = " " + keyword.group() if space else keyword.group()
                query = query.replace(token, "")
            except AttributeError:
                pass
        return query

    @staticmethod
    def get_operator_and_value(query, operator) -> re.Match or None:
        """Search for an operator and its following value"""
        return re.search(rf'{operator.lstrip(" ")}(\S+)', query)

    @staticmethod
    def get_operator_value(match_group) -> str:
        """Extract the value from an operator"""
        return match_group.split(":")[1].strip().rstrip('"')

    @staticmethod
    def convert_keyword(query, operator, keyword) -> str:
        """Replace an operator with its corresponding one"""
        return query.replace(operator, keyword)

    @staticmethod
    def days_since_epoch(date: datetime) -> int:
        """Return the number of days from the Unix epoch"""
        epoch_date = datetime(1970, 1, 1)
        difference = date - epoch_date
        return difference.days

    @staticmethod
    def validate_date(date_str) -> str or None:
        """Check if a date is formatted like YYYY-MM-DD"""
        error_message = "before: and after: format should be YYYY-MM-DD like 2025-01-01 or 2025-01-31."
        try:
            date = datetime.strptime(date_str, "%Y-%m-%d")
            if not (1 <= date.month <= 12 and 1 <= date.day <= 31):
                return error_message
        except ValueError:
            return error_message
        return None

    @staticmethod
    def format_date(date_str) -> str:
        """Format the date string into YYYY-MM-DD format if given without the 0"""
        date = datetime.strptime(date_str, "%Y-%m-%d")
        formatted_date = date.strftime("%Y-%m-%d")
        return formatted_date

    @staticmethod
    def check_in_quotation(query, operator) -> bool:
        """Check if the operator is within quotation marks"""
        return re.search(rf'"[^"]*{re.escape(operator)}[^"]*"', query) is not None

    @staticmethod
    def revert_domain(domain: str) -> str:
        """Revert domain name notation"""
        domain = domain.split(".")
        domain.reverse()
        return ".".join(domain)

    @staticmethod
    def check_last_value(value: str) -> bool:
        """Check if the last: operator value is valid (for last:)"""
        return bool(re.fullmatch(r"\d+[smhSMH]", value))

    def convert(self) -> dict:
        """Convert the operators in the query to the corresponding search engine URLs"""
        converted_queries = []

        for engine in self.engines:
            operators = engine["operators"]
            url = engine["url"]
            query = self.query

            # Handle dates
            if any(op in query for op in ["before:", "after:"]):
                date_handler = self.HandleDates(self, query, engine)
                if hasattr(date_handler, "error"):
                    return {"status": "error", "error": date_handler.error}
                query, chunk_result = date_handler.process_dates()
                if chunk_result:
                    url += chunk_result

            # Handle main operators
            for operator, details in operators.items():
                if operator != "*" and self.check_in_quotation(query, operator):
                    continue
                if operator in query:
                    # keword switch
                    if details["type"] == "keyword":
                        query = self.convert_keyword(
                            query, operator, details["keyword"]
                        )
                    # handle url type
                    elif details["type"] == "url":
                        if match := self.get_operator_and_value(query, operator):
                            operator_value = self.get_operator_value(match.group())
                            if engine["name"] == "Google" and operator == "country:":
                                operator_value = operator_value.upper()
                            chunk = details["chunk"].format(operator_value)
                            query = self.delete_keywords(query, [match])
                            url += chunk
                    # handle a split if * is not supported
                    elif details["type"] == "wildcard" and details["action"] == "split":
                        if matches := self.wildcard_pattern.findall(query):
                            for match in matches:
                                query = self.delete_keyword(query, match).strip()
                                splits = match.split("*")
                                converted_splits = [
                                    f'"{split.strip().replace('"', "").strip()}"'
                                    for split in splits
                                    if split != " "
                                ]
                                converted_splits = [
                                    split for split in converted_splits if split != '""'
                                ]
                                query += " " + " ".join(converted_splits)
                    # handle some Yandex special operators
                    elif details["type"] == "special" and engine["name"] == "Yandex":
                        operator_value = None
                        match = None
                        if operator == "country:":
                            yandex_countries = self.get_yandex_countries()
                            if match := self.get_operator_and_value(query, operator):
                                operator_value = self.get_operator_value(match.group())
                                operator_value = next(
                                    (
                                        country["yandex_code"]
                                        for country in yandex_countries
                                        if country["code"] == operator_value.upper()
                                    ),
                                    84,
                                )
                        elif operator == " lang:":
                            if match := self.get_operator_and_value(query, operator):
                                operator_value = self.get_operator_value(match.group())
                        chunk = details["chunk"].format(operator_value)
                        query = self.delete_keywords(query, [match])
                        url += chunk
                    # sub: handling
                    elif operator == "sub:" and details["type"] == "transform":
                        if match := self.get_operator_and_value(query, operator):
                            operator_value = self.get_operator_value(match.group())
                            if engine["name"] == "Google":
                                new_operator = f"site:*.{operator_value} -site:www.{operator_value}"
                                query = query.replace(match.group(), new_operator)
                            if engine["name"] == "Yandex":
                                rhost = self.revert_domain(operator_value)
                                new_operator = f"rhost:{rhost}.*"
                                query = query.replace(match.group(), new_operator)
                    # handle last: operator
                    elif operator == "last:" and details["type"] == "special":
                        if match := self.get_operator_and_value(query, operator):
                            operator_value = self.get_operator_value(match.group())
                            chunk = ""
                            if engine["name"] == "Google" and self.check_last_value(
                                operator_value
                            ):
                                if operator_value[-1].lower() == "s":
                                    lasting = operator_value[:-1]
                                    chunk = f"&tbs=qdr:s{lasting}"
                                if operator_value[-1].lower() == "m":
                                    lasting = operator_value[:-1]
                                    chunk = f"&tbs=qdr:n{lasting}"
                                if operator_value[-1].lower() == "h":
                                    lasting = operator_value[:-1]
                                    chunk = f"&tbs=qdr:h{lasting}"
                                query = self.delete_keywords(query, [match])
                                before_match = self.get_operator_and_value(
                                    query, "before:"
                                )
                                query = self.delete_keywords(query, [before_match])
                                after_match = self.get_operator_and_value(
                                    query, "after:"
                                )
                                query = self.delete_keywords(query, [after_match])
                                url += chunk
                    # handle !IP: operator
                    elif operator == "!IP" and details["type"] == "special":
                        ip_queries = self.ip_google(query)
                        for ip_query in ip_queries:
                            ip_query = ip_query.replace(" !IP", "").replace("!IP ", "").replace(" !IP ", "")
                            # ip_query = urllib.parse.quote(ip_query)
                            converted_queries.append(url.format(ip_query))
                    # delete unsupported operator
                    elif details["type"] == "delete":
                        if match := self.get_operator_and_value(query, operator):
                            space = True if engine["name"] == "Baidu" else False
                            query = self.delete_keywords(query, [match], space=space)
                    # handle Yandex for AND
                    elif engine["name"] == "Yandex" and details["type"] == "rewrite":
                        if operator == " AND ":
                            query = self.convert_and_to_yandex(query)

            # Handle DuckDuckGo countries, with multiple languages support per country
            if engine["name"] == "DuckDuckGo" and " country:" in query:
                if match := self.get_operator_and_value(query, "country:"):
                    operator_value = self.get_operator_value(match.group())
                    with open(
                        f"{json_data_dir}/duckduckgo_countries.json",
                        "r",
                        encoding="utf-8",
                    ) as f:
                        duckduckgo_countries = json.load(f)
                    regions = next(
                        (
                            country[operator_value]
                            for country in duckduckgo_countries
                            if operator_value in country
                        ),
                        ["wt-wt"],
                    )
                    query = self.delete_keywords(query, [match], space=True)
                    chunk = engine["operators"]["country:"]["chunk"]
                    for region in regions:
                        region_url = f"{url}{chunk.format(region)}"
                        converted_queries.append(region_url.format(query))
            else:
                # URL encode the query except for Baidu
                if engine["name"] != "Baidu":
                    query = urllib.parse.quote(query)
                converted_queries.append(url.format(query))
                # Remove the non IP search with Google
                if engine["name"] == "Google" and "!IP" in self.query:
                    converted_queries = converted_queries[:-1]

        converted_queries.reverse()
        return {"status": "ok", "urls": converted_queries}

    # Adapted from https://github.com/SeifElsallamy/gip by Seif Elsallamy
    @staticmethod
    def ip_google(search):
        o = 32 - len(search.split())
        template = "site:*.*.{}.*"
        ip_queries = []
        step = o
        start = 0
        while start < 256:
            end = min(start + step, 256)
            parts = [template.format(ii) for ii in range(start, end)]
            query = " |".join(parts)
            query = f"{search} ({query})"
            ip_queries.append(query)
            start = end
        return ip_queries

    @staticmethod
    def convert_and_to_yandex(query: str) -> str:
        """Converts AND to Yandex +."""
        # Define operators that should not be prefixed with '+'
        special_operators = {
            "site:",
            "intitle:",
            "inurl:",
            "intext:",
            "filetype:",
            "ext:",
            "sub:",
            "rhost:",
            "last:",
        }

        # Replace 'AND' with space to keep implicit AND behavior
        query = re.sub(r"\bAND\b", " ", query)

        # Tokenize input while keeping quoted phrases, operators, parentheses, and '|'
        pattern = (
            r"(\(|\)|"
            r"-?\b(?:" + "|".join(special_operators) + r")[^\s]+|"
            r'"[^"]+"|'
            r"-?\b\w+\b|"
            r"\|)"
        )
        tokens = re.findall(pattern, query)

        yandex_dork = []
        inside_parentheses = False
        open_paren_count = 0

        for token in tokens:
            if token == "(":
                # Opening parenthesis
                yandex_dork.append(token)
                inside_parentheses = True
                open_paren_count += 1
            elif token == ")":
                # Closing parenthesis
                yandex_dork.append(token)
                open_paren_count -= 1
                # If we've closed all parentheses, switch off inside_parentheses
                if open_paren_count <= 0:
                    inside_parentheses = False
            elif token == "|":
                # Pipe operator with single spacing
                yandex_dork.append(f" {token} ")
            elif token.startswith("-") or any(
                token.startswith(op) for op in special_operators
            ):
                # Operators or negations untouched
                yandex_dork.append(token)
            elif token.startswith('"'):
                # Quoted string untouched
                yandex_dork.append(token)
            else:
                # Normal word outside parentheses gets '+'
                if inside_parentheses:
                    # Inside parentheses, keep it without '+'
                    yandex_dork.append(token)
                else:
                    yandex_dork.append(f"+{token}")

        # Close  open parentheses
        while open_paren_count > 0:
            yandex_dork.append(")")
            open_paren_count -= 1

        # Clean up spacing
        result = " ".join(yandex_dork)
        # Remove double spaces from inserting ' | ',...
        result = re.sub(r"\s{2,}", " ", result).strip()

        return result

    class HandleDates:
        """Subclass to handle dates per search engine"""

        def __init__(self, search_instance, query, engine):
            self.search_instance = search_instance
            self.query = query
            self.engine = engine
            self.before = None
            self.after = None
            self.before_parenthesis = False
            self.after_parenthesis = False
            self.before_dt = None
            self.after_dt = None
            self.before_dt_s = None
            self.after_dt_s = None
            self.chunk = engine.get("date_chunk")
            self.today = datetime.now().strftime("%Y-%m-%d")

            self.before_kv = self.search_instance.get_operator_and_value(
                query, "before:"
            )
            self.after_kv = self.search_instance.get_operator_and_value(query, "after:")

            if self.before_kv:
                self.before = self.search_instance.get_operator_value(
                    self.before_kv.group()
                )
                validation = self.search_instance.validate_date(self.before)
                if validation:
                    self.error = validation
                    return
                self.before = self.search_instance.format_date(self.before)
                self.before_dt = datetime.strptime(self.before, "%Y-%m-%d")
                self.before_dt_s = self.before_dt.strftime("%Y%m%d")

            if self.after_kv:
                self.after = self.search_instance.get_operator_value(
                    self.after_kv.group()
                )
                validation = self.search_instance.validate_date(self.after)
                if validation:
                    self.error = validation
                    return
                self.after = self.search_instance.format_date(self.after)
                self.after_dt = datetime.strptime(self.after, "%Y-%m-%d")
                self.after_dt_s = self.after_dt.strftime("%Y%m%d")

        def process_dates(self) -> tuple:
            """Map date methods per engine name"""
            engine_handlers = {
                "Google": lambda: self.handle_google(),
                "Bing": lambda: self.handle_bing(),
                "Yandex": lambda: self.handle_yandex(),
                "Brave": lambda: self.handle_brave(),
                "DuckDuckGo": lambda: self.handle_duckduckgo(),
                "Baidu": lambda: self.handle_baidu(),
            }
            # Execute the appropriate handler method
            if self.engine["name"] in engine_handlers:
                return engine_handlers[self.engine["name"]]()
            return self.query, None

        def handle_google(self) -> tuple:
            """Handle Google date"""
            if self.before_kv:
                self.query = self.query.replace(
                    self.before_kv.group(), f"before:{self.before}"
                )
            if self.after_kv:
                self.query = self.query.replace(
                    self.after_kv.group(), f"after:{self.after}"
                )
            return self.query, None

        def handle_bing(self) -> tuple:
            """Handle Bing date"""
            before_days = (
                self.search_instance.days_since_epoch(self.before_dt)
                if self.before
                else self.search_instance.days_since_epoch(datetime.now())
            )
            after_days = (
                self.search_instance.days_since_epoch(self.after_dt)
                if self.after
                else 0
            )
            self.query = self.search_instance.delete_keywords(
                self.query, [self.before_kv, self.after_kv]
            )
            return self.query, self.chunk.format(after_days, before_days)

        def handle_yandex(self) -> tuple:
            """Handle Yandex date"""
            if self.before and self.after:
                if self.before == self.after:
                    self.query = self.query.replace(
                        self.before_kv.group(), f"date:{self.before_dt_s}"
                    )
                else:
                    self.query = self.query.replace(
                        self.before_kv.group(),
                        f"date:{self.after_dt_s}..{self.before_dt_s}",
                    )
            elif self.before and not self.after:
                self.query = self.query.replace(
                    self.before_kv.group(), f"date:<={self.before_dt_s}"
                )
            elif self.after and not self.before:
                self.query = self.query.replace(
                    self.after_kv.group(), f"date:>={self.after_dt_s}"
                )
            self.query = self.search_instance.delete_keywords(
                self.query, [self.before_kv, self.after_kv]
            )
            return self.query, None

        def handle_brave(self) -> tuple:
            """Handle Brave date"""
            if self.before and not self.after:
                self.after = datetime(1970, 1, 1).strftime("%Y-%m-%d")
            elif self.after and not self.before:
                self.before = self.today
            self.query = self.search_instance.delete_keywords(
                self.query, [self.before_kv, self.after_kv]
            )
            return self.query, self.chunk.format(self.after, self.before)

        def handle_duckduckgo(self) -> tuple:
            """Handle DuckDuckGo date"""
            if self.before and not self.after:
                self.after = datetime(1970, 1, 2).strftime("%Y-%m-%d")
            elif self.after and not self.before:
                self.before = self.today
            self.query = self.search_instance.delete_keywords(
                self.query, [self.before_kv, self.after_kv]
            )
            return self.query, self.chunk.format(self.after, self.before)

        def handle_baidu(self) -> tuple:
            """Handle Baidu date"""
            before_ts = (
                int(self.before_dt.timestamp())
                if self.before
                else int(datetime.now().timestamp())
            )
            after_ts = int(self.after_dt.timestamp()) if self.after else 0
            self.query = self.search_instance.delete_keywords(
                self.query, [self.before_kv, self.after_kv]
            )
            return self.query, self.chunk.format(after_ts, before_ts)



from nicegui import ui, app

from dorky.content import elements


def help_block() -> None:
    help_style = "border-radius: 50%; background-color: #c2c2c2; color: white; width: 20px; height: 20px; text-align: center; margin-right: 7px;"

    elements.ToggleButton("Help", color="#e6e6e6").on_click(
        lambda: app.storage.user.update(
            view_help=False if app.storage.user["view_help"] is True else True
        )
    )

    with ui.column().bind_visibility_from(app.storage.user, "view_help"):
        with ui.list().props("dense").style("color: #555;"):
            with ui.item():
                ui.label("1").style(help_style)
                ui.label(
                    "Authorize pop-up windows for Dorky website in your browser settings."
                )
            with ui.item():
                ui.label("2").style(help_style)
                ui.label("Choose the search engines. Highlighted = active.")
            with ui.item():
                ui.label("3").style(help_style)
                ui.label("Write your search query.")
            with ui.item():
                ui.label("4").style(help_style)
                ui.label("Press Enter or click the search button.")
